import yaml
import os
from typing import Dict, List, Any
from tag_creator.arguments import args
from tag_creator.logger import logger

CONFIGURATION: Dict[str, str] = {}


def read_configuration() -> Dict[str, Any]:
    global CONFIGURATION
    if not CONFIGURATION:
        with open(f"{os.path.dirname(__file__)}/configuration.yml", "r") as f:
            CONFIGURATION = yaml.safe_load(f)
            CONFIGURATION.update(custom_config())
    return CONFIGURATION


def custom_config() -> Dict[str, Any]:
    if os.path.exists(os.path.abspath(args.config)) and args.config:
        with open(args.config, "r") as f:
            logger.info("Custom config read!")
            return dict(yaml.safe_load(f))
    return {}


def allowed_commit_types() -> List[str]:
    cfg = read_configuration()
    return list(cfg["commit_types"])



import pytest
from unittest.mock import patch, MagicMock, call
from tag_creator.repository.version import ProjectVersionUpdater


@pytest.fixture(autouse=True)
def mock_git():
    with patch('tag_creator.repository.git.Git') as MockGit:
        yield MockGit

@pytest.fixture()
def mock_shell():
    def mock_exec(command, *args, **kwargs):
        if "tag --merged" in command:
            return "1.0.0"
        elif "log" in command:
            return "fix: some fix"
        elif "-n 1 HEAD" in command:
            return "1"
        else:
            return "foo-bar"

    with patch('tag_creator.utils.shell.exec', MagicMock()) as mock:
        mock.side_effect = mock_exec
        yield mock

def test_version_updated(mock_shell):
    ProjectVersionUpdater(repo_dir="fake_repo", release_branch="main", dry_run=False).create_new_verion()

    mock_shell.assert_has_calls([
        call("git -C fake_repo tag --merged 'main' --sort=creatordate --list '[0-9]*\\.[0-9]*\\.[0-9]*'"),
        call("git -C fake_repo rev-list -n 1 HEAD"),
        call("git -C fake_repo rev-list -n 1 1.0.0"),
        call("git -C fake_repo log -n 1 --pretty=%B"),
        call("git -C fake_repo tag -a '1.0.1' -m 'Automatically created tag'"),
        call("git -C fake_repo push origin 1.0.1"),
    ])

def test_dry_run_mode(mock_shell):
    ProjectVersionUpdater(repo_dir="fake_repo", release_branch="main", dry_run=True).create_new_verion()

    assert call("git -C fake_repo push origin 1.0.1") not in mock_shell.call_args_list

def test_tag_creation_with_prefix(mock_shell):
    ProjectVersionUpdater(repo_dir="fake_repo", release_branch="main", prefix="v").create_new_verion()

    mock_shell.assert_has_calls([
        call("git -C fake_repo tag --merged 'main' --sort=creatordate --list 'v[0-9]*\\.[0-9]*\\.[0-9]*'"),
        call("git -C fake_repo tag -a 'v1.0.1' -m 'Automatically created tag'"),
    ], any_order=True)

def test_not_allowed_type_raise():
    with pytest.raises(Exception, match="Can not determine commit majority based on it's message!"):
        with patch("tag_creator.utils.shell.exec", MagicMock()) as shell_mock:
            def mock_exec(command, *args, **kwargs):
                if "log" in command:
                    return "foo: bar"
                elif "tag --merged" in command:
                    return "1.0.0"
                elif "-n 1 HEAD" in command:
                    return "1"
                else:
                    return "foo-bar"
            shell_mock.side_effect = mock_exec
            ProjectVersionUpdater(repo_dir="fake_repo", release_branch="main", prefix="v").create_new_verion()

@pytest.mark.parametrize(
    "commit_msg, increments",
    [
        ("fix: foo", (0, 0, 1)),
        ("perf: docs", (0, 0, 1)),
        ("test: refactor", (0, 0, 1)),
        ("feat: new feature", (0, 1, 0)),
        ("style: new feature", (0, 0, 1)),
        ("chore: new feature", (0, 0, 1)),
        ("ci: new feature", (0, 1, 0)),
        ("build: new feature", (0, 1, 0)),
        ("revert: new feature", (0, 1, 0)),
        ("feat!: test", (1, 0, 0)),
        ("BREAKING_CHANGE: foo", (1, 0, 0)),
        ("fix: some fix\nBREAKING_CHANGE: foo", (1, 0 , 0))
    ],
)
def test_increments_based_on_commit_msg(commit_msg, increments):
    with patch("tag_creator.utils.shell.exec", MagicMock()) as shell_mock:
        def mock_exec(command, *args, **kwargs):
            if "log" in command:
                return commit_msg
            elif "tag --merged" in command:
                return "1.0.0"
            elif "-n 1 HEAD" in command:
                return "1"
            else:
                return "foo-bar"

        shell_mock.side_effect = mock_exec

        ProjectVersionUpdater(repo_dir="fake_repo", release_branch="main", prefix="v").create_new_verion()
        expected_tag = f"v{1 + increments[0]}.{0 + increments[1]}.{0 + increments[2]}"

        shell_mock.assert_any_call(f"git -C fake_repo tag -a '{expected_tag}' -m 'Automatically created tag'")



import re
import tag_creator.configuration as cfg
from tag_creator.logger import logger
from tag_creator.repository.git import Git
from typing import List


class ProjectVersionUpdater(Git):

    MAJOR_VER = "major"
    MINOR_VER = "minor"
    PATCH_VER = "patch"

    def __init__(self, repo_dir: str, release_branch: str, dry_run: bool = False, prefix: str = "") -> None:
        super().__init__(repo_dir=repo_dir)
        self.release_branch = release_branch
        self.prefix = prefix
        self.dry_run = dry_run

    def current_tag(self) -> str:
        tags = self.__all_tags()
        if not tags:
            raise Exception("There is no initial tag!")
        return str(tags[-1])

    def create_new_verion(self) -> None:
        """Create new tag for the release branch.
        Tag pattern: ${prefix}d.d.d .Initial tag must be created manually.
        """
        current_version = self.current_tag()
        logger.info(f"Current version is: {current_version}")
        if self.__is_tag_on_current_head(current_version):
            logger.warning(
                f"There are no new changes starting from the latest tag: {current_version}. Skip tag creation."
            )
            return

        version_without_prefix = self.__version_witout_prefix(current_version)
        new_tag = self.__add_prefix(".".join(map(
            str,
            self.__increment_version(version_without_prefix, self.log("-n 1 --pretty=%B")))
        ))

        self.__create_tag(new_tag, "Automatically created tag")
        if self.dry_run:
            logger.info("Dry run! New tag will not be pushed.")
            return
        self.push(new_tag)

    def __all_tags(self) -> list[str]:
        return list(
            self.tag(
                f"--merged '{self.release_branch}' --sort=creatordate --list "
                f"'{self.prefix}[0-9]*\\.[0-9]*\\.[0-9]*'"
            ).strip().splitlines()
        )

    def __commit_hash(self, ref: str) -> str:
        return str(self.rev_list(f"-n 1 {ref}"))

    def __is_tag_on_current_head(self, tag: str) -> bool:
        head_hash = self.__commit_hash("HEAD")
        tag_hash = self.__commit_hash(tag)

        return (head_hash == tag_hash)

    def __create_tag(self, tag: str, msg: str = "") -> None:
        logger.info(f"New tag will be created and pushed. Tag: {tag}, message {msg}")
        self.tag(f"-a '{tag}' -m '{msg}'")

    def __increment_version(self, version: str, change_title: str) -> tuple[int, int, int]:
        major, minor, patch = map(int, version.split('.'))
        new_major, new_minor, new_patch = self.__patch_increment_based_on_title(change_title)

        if new_major > 0:
            major += new_major
            minor = 0
            patch = 0
        elif new_minor > 0:
            minor += new_minor
            patch = 0
        else:
            patch += new_patch

        return (major, minor, patch)

    def __version_witout_prefix(self, version: str) -> str:
        return version.lstrip(self.prefix)

    def __add_prefix(self, version: str) -> str:
        return f"{self.prefix}{version}"

    def __patch_increment_based_on_title(self, title: str) -> tuple[int, int, int]:
        """Extract patch based on the change title.

        Args:
            title (_type_): merge request or commit.

        Raises:
            Exception: There are no allowed commit types in the configuration.

        Returns:
            tuple: major, minor and patch increment
        """
        logger.info(f"Determine increments for new tag based on commit (MR) msg: {title}")

        if self.__is_major_version(title):
            return 1, 0, 0
        elif self.__is_starts_from_version(self.MINOR_VER, title):
            return 0, 1, 0
        elif self.__is_starts_from_version(self.PATCH_VER, title):
            return 0, 0, 1
        else:
            raise Exception("Can not determine commit majority based on it's message!")

    def __is_major_version(self, commit_msg: str) -> bool:
        return bool(
            re.match("^(" + ('|').join(self.__all_commit_types()) + ")[a-z()]*!:", commit_msg, re.MULTILINE) or
            re.search("^(" + ('|').join(self.__types_for_majority(self.MAJOR_VER)) + r"): \w", commit_msg, re.MULTILINE)
        )

    def __types_for_majority(self, majority: str) -> List[str]:
        types: list[str] = []
        types.extend(
            [entry["type"] for entry in cfg.allowed_commit_types() if entry["majority"] == majority]  # type: ignore
        )
        return [item for sublist in types for item in sublist]

    def __all_commit_types(self) -> list[str]:
        majority: list[str] = []
        for entry in cfg.allowed_commit_types():
            if isinstance(entry, dict) and "type" in entry:
                majority.extend(entry["type"])
            else:
                logger.warning(f"Skipping invalid entry: {entry}")

        return majority

    def __is_starts_from_version(self, version: str, commit_msg: str) -> bool:
        return bool(re.match(
            "^(" + "|".join(self.__types_for_majority(version)) + ")[a-z()]*:",
            commit_msg,
            re.MULTILINE
        ))



from tag_creator.utils import shell


class Git():

    def __init__(self, repo_dir: str = ".") -> None:
        self.repo_dir = repo_dir

    def tag(self, args: str = "") -> str:
        return self.__git_cmd(f"tag {args}")

    def rev_list(self, args: str = "") -> str:
        return self.__git_cmd(f"rev-list {args}")

    def push(self, reference: str = "") -> str:
        return self.__git_cmd(f"push origin {reference}")

    def log(self, args: str = "") -> str:
        return self.__git_cmd(f"log {args}")

    def __git_cmd(self, cmd: str) -> str:
        return shell.exec(f"git -C {self.repo_dir} {cmd}")



import subprocess
from tag_creator.logger import logger
from typing import Any


def exec(command: str, *args: tuple[str, ...], **kwargs: Any) -> str:
    """Execute shell command

    Args:
        command (str): cmd to execute.

    Returns:
        str: command stdout
    """
    try:
        result = subprocess.run(
            command, shell=True, check=True, capture_output=True, encoding="UTF-8", *args, **kwargs  # type: ignore
        )
        return str(result.stdout)
    except subprocess.CalledProcessError as ex:
        logger.error(f"Execution failed. Status code: {ex.returncode}. Message: {ex.stderr}")
        raise



import argparse


def parse() -> argparse.Namespace:
    parser = argparse.ArgumentParser(prog="Repository tool.")
    parser.add_argument("--repo_dir", help="Repository root directory", default=".")
    parser.add_argument("--release_branch", help="Release branch", required=False, default="main")
    parser.add_argument("--create_new_tag", help="Update tag", action="store_true", default=False)
    parser.add_argument("--dry_run", help="Do not update remote", action="store_true", default=False)
    parser.add_argument("--show_config", help="Show default config", action="store_true", default=False)
    parser.add_argument("--config", help="Config file", required=False, default="")
    parser.add_argument("--current_version", help="Show current version", action="store_true", default=False)
    parser.add_argument("--tag_prefix", help="Prefix to version number", required=False, default="")

    return parser.parse_args()


args = parse()



import yaml
from tag_creator.repository.version import ProjectVersionUpdater
from tag_creator.logger import logger
from tag_creator import configuration as cfg
from tag_creator.arguments import args


if __name__ == "__main__":

    if args.show_config:
        logger.info(yaml.dump(cfg.read_configuration()))

    pvu = ProjectVersionUpdater(
        repo_dir=args.repo_dir,
        release_branch=args.release_branch,
        prefix=args.tag_prefix,
        dry_run=args.dry_run
    )

    if args.create_new_tag:
        pvu.create_new_verion()

    if args.current_version:
        logger.info(f"Current tag: {pvu.current_tag()} Branch: {args.release_branch}")
